-----------------------------------------------------------------------
--
--  File:        vgagraph.adb
--  Description: Basic VGA graphics
--  Rev:         0.4
--  Date:        16/02/96
--  Author:      Jerry van Dijk
--  Mail:        jerry@jvdsys.nextjk.stuyts.nl
--
--  Copyright (c) Jerry van Dijk, 1996
--  Forelstraat 211
--  2037 KV  HAARLEM
--  THE NETHERLANDS
--  tel int + 31 23 540 1052
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------
--
--  This package provides basic VGA graphics capabilities for GNAT/DOS.
--  It is currently based on GNAT 3.01 / DJGPP 1.12m4 running under
--  both vcpi and dpmi. The package does not depend on any djgpp
--  graphics functions as it is /not/ a C library binding but a new
--  implementation written completely in Ada95 itself.
--  As it was written as a learning Ada95 program it is in its present
--  form probably not very portable to other enviroments.
--  The choice of what functionality to provide was based the expected
--  needs of a user-interface package (probably my next Ada95 program,
--  incorporating the new OO features).
--
-----------------------------------------------------------------------

with Ada.Unchecked_Deallocation;

with DJGPP_Library; use DJGPP_Library;

package body VGA_Graphics is

--------------------
--  PRIVATE PART  --
--------------------

   -----------------
   --  Constants  --
   -----------------

   --  VGA screen constants
   Video_Base_address           : constant unsigned_32 := 16#A0000#;
   Bytes_Per_Line               : constant unsigned_32 := 16#00000050#;

   --  BIOS constants
   BIOS_VGA_Mode                : constant unsigned_8  := 16#12#;
   BIOS_Video_Interrupt         : constant unsigned_16 := 16#10#;

   --  VGA controller values
   VGA_Set_Color_Register       : constant unsigned_8  := 0;
   VGA_Set_Write_Plane_Register : constant unsigned_8  := 1;
   VGA_Sequential_Register      : constant unsigned_8  := 2;
   VGA_Put_Rule_Register        : constant unsigned_8  := 3;
   VGA_Set_Read_Plane_Register  : constant unsigned_8  := 4;
   VGA_Set_Bitmask_Register     : constant unsigned_8  := 8;
   VGA_All_Planes               : constant unsigned_8  := 16#FF#;

   --  VGA controller adresses
   Vga_Seqential_Index          : constant unsigned_16 := 16#03C4#;
   Vga_Seqential                : constant unsigned_16 := 16#03C5#;
   Vga_Graphics_Index           : constant unsigned_16 := 16#03CE#;
   Vga_Graphics                 : constant unsigned_16 := 16#03CF#;

   --  Default palette colors
   Default_Palette : constant Palette := (16#00#, 16#01#, 16#02#, 16#03#,
                                          16#04#, 16#05#, 16#14#, 16#07#,
                                          16#38#, 16#39#, 16#3A#, 16#3B#,
                                          16#3C#, 16#3D#, 16#3E#, 16#3F#,
                                          16#00#);

   ----------------
   --  Font Type --
   ----------------
   Font_Max_Char   : constant Natural := 255;
   Font_Max_Height : constant Natural := Font_Height - 1;

   type Font_Char is array (0..Font_Max_Height) of unsigned_8;
   type Font_Type is array (0..Font_Max_Char) of Font_Char;

   ------------------------
   --  Global Variables  --
   ------------------------

   --  Package status
   VGA_Status       : Boolean := False;

   --  Keep original mode when switching to VGA
   Original_Mode    : unsigned_8;

   --  Keep original palette when switchin to VGA
   Original_Palette : Palette;

   --  The system font
   System_Font      : Font_Type;

   ---------------------------------
   --  Set VGA Graphics Register  --
   ---------------------------------
   procedure Set_VGA_Graphics_Register(Register : in unsigned_8;
                                       Value    : in unsigned_8) is
   begin
      outportb(VGA_Graphics_Index, Register);
      outportb(VGA_Graphics, Value);
   end Set_VGA_Graphics_Register;
   pragma Inline(Set_VGA_Graphics_Register);

   --------------------------------
   --  VGA Set Sequential Plane  --
   --------------------------------
   procedure VGA_Set_Sequential_Plane(Plane : in unsigned_8) is
   begin
      outportb(Vga_Seqential_Index, VGA_Sequential_Register);
      outportb(Vga_Seqential, Plane);
   end VGA_Set_Sequential_Plane;
   pragma Inline(VGA_Set_Sequential_Plane);

   -------------------------------------
   --  VGA Get Byte From Screen Memory  --
   -------------------------------------
   function VGA_Get_Byte(Address : in unsigned_32)
                         return unsigned_8 is
      Selector : constant unsigned_16 := go32_conventional_mem_selector;
   begin
      return farpeekb(Selector, Address);
   end VGA_Get_Byte;
   pragma Inline(VGA_Get_Byte);

   -------------------------------------
   --  VGA Put Byte In Screen Memory  --
   -------------------------------------
   procedure VGA_Put_Byte(Address : in unsigned_32;
                          Value   : in unsigned_8) is
      Dummy    : unsigned_8; -- To get byte into latch
      Selector : constant unsigned_16 := go32_conventional_mem_selector;
   begin
      Dummy := VGA_Get_Byte(Address);
      farpokeb(Selector, Address, Value);
   end VGA_Put_Byte;
   pragma Inline(VGA_Put_Byte);

   ------------------------------------------------------------
   --  VGA Get Byte From Screen Memory With Defaul Selector  --
   ------------------------------------------------------------
   function VGA_Get_NS_Byte(Address : in unsigned_32)
                         return unsigned_8 is
   begin
      return farnspeekb(Address);
   end VGA_Get_NS_Byte;
   pragma Inline(VGA_Get_NS_Byte);

   ---------------------------------------------------------
   --  VGA Put Byte In Screen Memory With Defaul Selector --
   ---------------------------------------------------------
   procedure VGA_Put_NS_Byte(Address : in unsigned_32;
                             Value   : in unsigned_8) is
      Dummy : unsigned_8; -- To get byte into latch
   begin
      Dummy := VGA_Get_NS_Byte(Address);
      farnspokeb(Address, Value);
   end VGA_Put_NS_Byte;
   pragma Inline(VGA_Put_NS_Byte);

   -------------------------------
   --  VGA Set Screen Selector  --
   -------------------------------
   procedure VGA_Set_Screen_Selector is
      Selector  : constant unsigned_16 := go32_conventional_mem_selector;
   begin
      farsetsel(Selector);
   end VGA_Set_Screen_Selector;
   pragma Inline(VGA_Set_Screen_Selector);

   --------------------------
   --  VGA Set Draw Color  --
   --------------------------
   procedure VGA_Set_Color(Color : in Screen_Color) is
   begin
      Set_VGA_Graphics_Register(VGA_Set_Color_Register,
                                Screen_Color'Pos(Color));
   end VGA_Set_Color;
   pragma Inline(VGA_Set_Color);

   ---------------------------
   --  VGA Set Read Plane  --
   ---------------------------
   procedure VGA_Set_Read_Plane(Plane : in unsigned_8) is
   begin
      Set_VGA_Graphics_Register(VGA_Set_Read_Plane_Register, Plane);
   end VGA_Set_Read_Plane;
   pragma Inline(VGA_Set_Read_Plane);

   ---------------------------
   --  VGA Set Write Plane  --
   ---------------------------
   procedure VGA_Set_Write_Plane(Plane : in unsigned_8) is
   begin
      Set_VGA_Graphics_Register(VGA_Set_Write_Plane_Register, Plane);
   end VGA_Set_Write_Plane;
   pragma Inline(VGA_Set_Write_Plane);

   -----------------------
   --  VGA Set Bitmask  --
   -----------------------
   procedure VGA_Set_Bitmask(Mask : in unsigned_8) is
   begin
      Set_VGA_Graphics_Register(VGA_Set_Bitmask_Register, Mask);
   end VGA_Set_Bitmask;
   pragma Inline(VGA_Set_Bitmask);

   ------------------------
   --  VGA Set Put Rule  --
   ------------------------
   procedure VGA_Set_Put_Rule(Rule : in Put_Rule_Type) is
   begin
      Set_VGA_Graphics_Register(VGA_Put_Rule_Register,
        Put_Rule_Type'Enum_Rep(Rule));
   end VGA_Set_Put_Rule;
   pragma Inline(VGA_Set_Put_Rule);

   -------------------------------
   --  Calculate Pixel Bitmask  --
   -------------------------------
   function Calculate_Pixel_Bitmask(X : in Horizontal_Location)
                                    return unsigned_8 is
      Mask : constant unsigned_8 := 2#10000000#;
   begin
      return Shift_Right(Mask, X mod 8);
   end Calculate_Pixel_Bitmask;
   pragma Inline(Calculate_Pixel_Bitmask);

   -------------------------------
   --  Calculate Left Bytemask  --
   -------------------------------
   function Calculate_Left_Bytemask(X : in Horizontal_Location)
                                    return unsigned_8 is
      Mask : constant unsigned_8 := 2#11111111#;
   begin
      return Shift_Right(Mask, X mod 8);
   end Calculate_Left_Bytemask;
   pragma Inline(Calculate_Left_Bytemask);

   --------------------------------
   --  Calculate Right Bytemask  --
   --------------------------------
   function Calculate_Right_Bytemask(X : in Horizontal_Location)
                                     return unsigned_8 is
      Mask : constant unsigned_8 := 2#11111111#;
   begin
      return Shift_Left(Mask, 7 - (X mod 8));
   end Calculate_Right_Bytemask;
   pragma Inline(Calculate_Right_Bytemask);

   ------------------------------
   --  Calculate Byte Address  --
   ------------------------------
   function Calculate_Byte_Address(X : in Horizontal_Location;
                                   Y : in Vertical_location)
                                   return unsigned_32 is
   begin
      return Video_Base_Address +
             unsigned_32(Y) * Bytes_Per_Line +
             unsigned_32(X / 8);
   end Calculate_Byte_Address;
   pragma Inline(Calculate_Byte_Address);

   -------------------------
   --  Plot Preset Point  --
   -------------------------
   procedure VGA_Preset_Plot(X : in Integer;
                             Y : in Integer) is
   begin
      if X in 0..Horizontal_Maximum and
         Y in 0..Vertical_Maximum then
         VGA_Set_Bitmask(Calculate_Pixel_Bitmask(X));
         VGA_Put_NS_Byte(Calculate_Byte_Address(X, Y), 16#FF#);
      end if;
   end VGA_Preset_Plot;
   pragma Inline(VGA_Preset_Plot);

   ---------------------------------
   --  Return Current Video Mode  --
   ---------------------------------
   function Current_Video_Mode return unsigned_8 is
      Regs : go32_dpmi_registers;
   begin
      Regs.ax := 16#0F00#;
      Regs.ss := 0; Regs.sp := 0; Regs.flags := 0;
      go32_dpmi_simulate_int(BIOS_Video_Interrupt, Regs);
      return unsigned_8(Regs.ax and 16#00FF#);
   end Current_Video_Mode;

   --------------------------------
   --  Check That VGA is Active  --
   --------------------------------
   procedure Check_Proper_Mode is
   begin
      if VGA_Status = False then
         raise Not_In_VGA_Error;
      end if;
   end Check_Proper_Mode;
   pragma Inline(Check_Proper_Mode);

   -----------------------
   --  Set Video Video  --
   -----------------------
   procedure Set_Video_Mode(Mode : in unsigned_8) is
      Regs : go32_dpmi_registers;
   begin
      Regs.ax := unsigned_16(Mode);
      Regs.ss := 0; Regs.sp := 0; Regs.flags := 0;
      go32_dpmi_simulate_int(BIOS_Video_Interrupt, Regs);
   end Set_Video_Mode;

   ------------------------------
   --  Initialize System Font  --
   ------------------------------
   procedure Init_System_Font is
      Address : unsigned_32;
      Regs    : go32_dpmi_registers;
   begin
      Regs.ax := 16#1130#;
      Regs.bx := 16#0200#;
      Regs.ss := 0; Regs.sp := 0; Regs.flags := 0;
      go32_dpmi_simulate_int(BIOS_Video_Interrupt, Regs);
      Address := 16 * unsigned_32(Regs.es) + unsigned_32(Regs.bp);
      for Num in 0..Font_Max_Char loop
         For Byte in 0..Font_Max_Height loop
            System_Font(Num)(Byte) := VGA_Get_Byte(Address);
            Address := Address + 1;
         end loop;
      end loop;
   end Init_System_Font;

   --------------------------------
   --  Initialize Screen Buffer  --
   --------------------------------
   procedure Initialize(B : in out Screen_Buffer) is
      Horizontal_Bytes : Integer := B.Width / 8;
   begin
      if B.Width = 0 or B.Height = 0 then
         raise Screen_Buffer_Error;
      else
         if B.Width mod 8 /= 0 then
            Horizontal_Bytes := Horizontal_Bytes + 1;
         end if;
         B.Data := new Planes(1..(4 * Horizontal_Bytes * B.Height));
      end if;
   end Initialize;

   ------------------------------
   --  Finalize Screen Buffer  --
   ------------------------------
   procedure Finalize(B : in out Screen_Buffer) is
      procedure Free is new Ada.Unchecked_Deallocation(Planes, Planes_Access);
   begin
      Free(B.Data);
   end Finalize;

   ---------------------------
   --  Reset VGA registers  --
   ---------------------------
   procedure Reset_VGA_Registers is
   begin
      VGA_Set_Color(Screen_Color'Val(15));
      VGA_Set_Bitmask(VGA_All_Planes);
      VGA_Set_Write_Plane(0);
      VGA_Set_Read_Plane(VGA_All_Planes);
      VGA_Set_Sequential_Plane(16#0F#);
   end Reset_VGA_Registers;

-------------------
--  PUBLIC PART  --
-------------------

   --------------------------
   --  Switch to VGA Mode  --
   --------------------------
   procedure VGA_Mode is
   begin
      if VGA_Status = False then
         Original_Mode := Current_Video_Mode;
         Set_Video_Mode(BIOS_VGA_Mode);
         if Current_Video_Mode /= BIOS_VGA_Mode then
            raise No_VGA_Error;
         end if;
         VGA_Status := True;
         Original_Palette := Read_All_Palette;
         Write_All_Palette(Default_Palette);
         Init_System_Font;
      end if;
   end VGA_Mode;

   --------------------------
   --  Switch to Text Mode --
   --------------------------
   procedure TXT_Mode is
   begin
      Check_Proper_Mode;
      Write_All_Palette(Original_Palette);
      Reset_VGA_Registers;
      Set_Video_Mode(Original_Mode);
      if Current_Video_Mode /= Original_Mode then
         raise No_TXT_Mode_Error;
      end if;
      VGA_Status := False;
   end TXT_Mode;

   --------------------
   --  Clear Screen  --
   --------------------
   procedure Clear_Screen(Color : in Screen_Color := Black) is
      Dummy     : unsigned_8; -- To get byte into latch
      Address   : unsigned_32          := Video_Base_Address;
      Num_Bytes : Positive :=
                  (Horizontal_Maximum + 1) * (Vertical_Maximum + 1) / 8;
   begin
      Check_Proper_Mode;
      VGA_Set_Color(Color);
      VGA_Set_Bitmask(2#11111111#);
      VGA_Set_Write_Plane(VGA_All_Planes);
      VGA_Set_Screen_Selector;
      for I in 0..Num_Bytes loop
         Dummy := farnspeekb(Address);
         farnspokeb(Address, 16#FF#);
         Address := Address + 1;
      end loop;
      VGA_Set_Write_Plane(0);
   end Clear_Screen;

   ---------------------------
   --  Put Pixel On Screen  --
   ---------------------------
   procedure Put_Pixel(X_Pos : in Horizontal_Location;
                       Y_Pos : in Vertical_Location;
                       Color : Screen_Color) is
   begin
      Check_Proper_Mode;
      VGA_Set_Color(Color);
      VGA_Set_Write_Plane(VGA_All_Planes);
      VGA_Preset_Plot(X_Pos, Y_Pos);
      VGA_Set_Write_Plane(0);
   end Put_Pixel;

   -----------------------
   --  Get Pixel Color  --
   -----------------------
   function Get_Pixel(X_Pos : in Horizontal_Location;
                      Y_Pos : in Vertical_Location)
                      return Screen_Color is
      Byte     : unsigned_8;
      Color    : unsigned_8           := 0;
      Bit_Mask : constant unsigned_8  := Calculate_Pixel_Bitmask(X_Pos);
      Address  : constant unsigned_32 := Calculate_Byte_Address(X_Pos, Y_Pos);
   begin
      Check_Proper_Mode;
      VGA_Set_Bitmask(2#11111111#);
      for Plane in reverse 0..3 loop
         Color := Shift_Left(Color, 1);
         VGA_Set_Read_Plane(unsigned_8(Plane));
         Byte := VGA_Get_Byte(Address);
         if (Byte and Bit_Mask) /= 0 then
            Color := Color or 2#00000001#;
         end if;
      end loop;
      return Screen_Color'Val(Color);
   end Get_Pixel;

   ---------------------------
   --  Write Vertical Line  --
   ---------------------------
   procedure Vertical_Line(X_Pos  : in Horizontal_Location;
                           Top    : in Vertical_Location;
                           Bottom : in Vertical_Location;
                           Color  : in Screen_Color) is
      Address  : unsigned_32;
      Y1       : Vertical_Location   := Top;
      Y2       : Vertical_Location   := Bottom;
      Bit_Mask : constant unsigned_8 := Calculate_Pixel_bitmask(X_Pos);
   begin
      Check_Proper_Mode;
      if Top = Bottom then
         Put_Pixel(X_Pos, Top, Color);
      else
         if Top > Bottom then
            Y1 := Bottom;
            Y2 := Top;
         end if;
      end if;
      VGA_Set_Color(Color);
      VGA_Set_Write_Plane(VGA_All_Planes);
      VGA_Set_Bitmask(Bit_Mask);
      Address := Calculate_Byte_Address(X_Pos, Y1);
      VGA_Set_Screen_Selector;
      for Y in Y1..Y2 loop
         VGA_Put_NS_Byte(Address, 16#FF#);
         Address := Address + Bytes_Per_Line;
      end loop;
      VGA_Set_Write_Plane(0);
   end Vertical_Line;

   -----------------------------
   --  Write Horizontal Line  --
   -----------------------------
   procedure Horizontal_Line(Left  : in Horizontal_Location;
                             Right : in Horizontal_Location;
                             Y_Pos : in Vertical_Location;
                             Color : in Screen_Color) is
      Left_Mask  : unsigned_8;
      Right_Mask : unsigned_8;
      Address    : unsigned_32;
      X1         : Natural := Left;
      X2         : Natural := Right;
   begin
      Check_Proper_Mode;
      if Left = Right then
         Put_Pixel(Left, Y_Pos, Color);
      else
         If Left > Right then
            X1 := Right;
            X2 := Left;
         end if;
         VGA_Set_Color(Color);
         VGA_Set_Screen_Selector;
         VGA_Set_Write_Plane(VGA_All_Planes);
         Left_Mask  := Calculate_Left_Bytemask(X1);
         Right_Mask := Calculate_Right_Bytemask(X2);
         Address    := Calculate_Byte_Address(X1, Y_Pos);
         if (X1 / 8) = (X2 / 8) then
            -- X1 and X2 are within one byte
            VGA_Set_Bitmask(Left_Mask and Right_Mask);
            VGA_Put_NS_Byte(Address, 16#FF#);
         elsif (X1 / 8 + 1) = (X2 / 8) then
            -- X1 and X2 are adjectent bytes
            VGA_Set_Bitmask(Left_Mask);
            VGA_Put_NS_Byte(Address, 16#FF#);
            VGA_Set_Bitmask(Right_Mask);
            VGA_Put_NS_Byte(Address+1, 16#FF#);
         else
            -- X1 and X2 have full bytes in between
            VGA_Set_Bitmask(Left_Mask);
            VGA_Put_NS_Byte(Address, 16#FF#);
            VGA_Set_Bitmask(16#FF#);
            X1 := X1 - X1 mod 8 + 8;
            for I in (X1 / 8)..(X2 / 8 - 1) loop
               Address := Address + 1;
               VGA_Put_NS_Byte(Address, 16#FF#);
            end loop;
            VGA_Set_Bitmask(Right_Mask);
            VGA_Put_NS_Byte(Address+1, 16#FF#);
         end if;
         VGA_Set_Write_Plane(0);
      end if;
   end Horizontal_Line;

   -------------------
   --  Draw a line  --
   -------------------
   procedure Draw_Line(Left   : in Horizontal_Location;
                       Top    : in Vertical_Location;
                       Right  : in Horizontal_Location;
                       Bottom : in Vertical_Location;
                       Color  : in Screen_Color) is

      type Direction_Type is (Normal, Reversed);

      Step             : Integer;
      Step_X, Step_Y   : Integer;
      Delta_X, Delta_Y : Integer;
      Y1               : Integer := Top;
      X1               : Integer := Left;
      Direction        : Direction_Type := Normal;

      procedure Swap(A : in out Integer;
                     B : in out Integer) is
         Temp : constant Integer := A;
      begin
         A := B;
         B := Temp;
      end Swap;

   begin
      Check_Proper_Mode;
      VGA_Set_Color(Color);
      VGA_Set_Screen_Selector;
      Delta_X := abs (Right - X1);
      Delta_Y := abs (Bottom - Y1);
      VGA_Set_Write_Plane(VGA_All_Planes);
      if Right > X1 then
         Step_X := 1;
      else
         Step_X := -1;
      end if;
      if Bottom > Y1 then
         Step_Y := 1;
      else
         Step_Y := -1;
      end if;
      if Delta_Y > Delta_X then
         swap(X1, Y1);
         swap(Step_X, Step_Y);
         swap(Delta_X, Delta_Y);
         Direction := Reversed;
      end if;
      Step := 2 * Delta_Y - Delta_X;
      for I in 1..Delta_X loop
         if Direction = Reversed then
            VGA_Preset_Plot(Y1, X1);
         else
            VGA_Preset_Plot(X1, Y1);
         end if;
         while Step >= 0 loop
            Y1   := Y1   + Step_Y;
            Step := Step - 2 * Delta_X;
         end loop;
         X1   := X1   + Step_X;
         Step := Step + 2 * Delta_Y;
      end loop;
      VGA_Preset_Plot(Right, Bottom);
      VGA_Set_Write_Plane(0);
   end Draw_Line;

   ---------------------
   --  Draw a circle  --
   ---------------------
   procedure Draw_Circle(X_Pos  : in Horizontal_Location;
                         Y_Pos  : in Vertical_Location;
                         Radius : in Positive;
                         Color  : in Screen_Color) is
      X    : Integer := 0;
      Y    : Integer := Radius;
      Diff : Integer := 2 * (1 - Radius);
   begin
      Check_Proper_Mode;
      VGA_Set_Color(Color);
      VGA_Set_Screen_Selector;
      VGA_Set_Write_Plane(VGA_All_Planes);
      VGA_Preset_Plot(X_Pos + Radius, Y_Pos);
      VGA_Preset_Plot(X_Pos - Radius, Y_Pos);
      While Y > 0 loop
         VGA_Preset_Plot(X_Pos + X, Y_Pos + Y);
         VGA_Preset_Plot(X_Pos + X, Y_Pos - Y);
         VGA_Preset_Plot(X_Pos - X, Y_Pos + Y);
         VGA_Preset_Plot(X_Pos - X, Y_Pos - Y);
         if Diff + Y > 0 then
            Y    := Y - 1;
            Diff := Diff - 2 * Y - 1;
         end if;
         if X > Diff then
            X    := X + 1;
            Diff := Diff + 2 * X + 1;
         end if;
      end loop;
      VGA_Set_Write_Plane(0);
   end Draw_Circle;

   ------------------
   --  Fill a Box  --
   ------------------
   procedure Fill_Box(Left   : in Horizontal_Location;
                      Top    : in Vertical_Location;
                      Right  : in Horizontal_Location;
                      Bottom : in Vertical_Location;
                      Color  : in Screen_Color)is

      Left_Mask  : unsigned_8;
      Right_Mask : unsigned_8;
      Address    : unsigned_32;
      Y1         : Vertical_Location   := Top;
      X1         : Horizontal_Location := Left;
      X2         : Horizontal_Location := Right;
      Y2         : Vertical_Location   := Bottom;

      procedure Draw_Strip(Y1      : in Vertical_Location;
                           Y2      : in Vertical_Location;
                           Address : in unsigned_32) is
         Memory : unsigned_32 := Address;
      begin
         for Y in Y1..Y2 loop
            VGA_Put_NS_Byte(Memory, 16#FF#);
            Memory := Memory + Bytes_Per_Line;
         end loop;
      end Draw_Strip;
      Pragma Inline(Draw_Strip);

   begin
      Check_Proper_Mode;
      if Left = Right and Top = Bottom then
         Put_Pixel(Left, Top, Color);
      elsif Left = Right then
         Vertical_Line(Left, Top, Bottom, Color);
      elsif Top = Bottom then
         Horizontal_Line(Left, Right, Top, Color);
      else
         if Left > Right then
            X1 := Right;
            X2 := Left;
         end if;
         if Top > Bottom then
            Y1 := Bottom;
            Y2 := Top;
         end if;
         VGA_Set_Color(Color);
         VGA_Set_Screen_Selector;
         VGA_Set_Write_Plane(VGA_All_Planes);
         Left_Mask  := Calculate_Left_Bytemask(X1);
         Right_Mask := Calculate_Right_Bytemask(X2);
         Address    := Calculate_Byte_Address(X1, Y1);
         if (X1 / 8) = (X2 / 8) then
            -- X1 and X2 are within one byte
            VGA_Set_Bitmask(Left_Mask and Right_Mask);
            Draw_Strip(Y1, Y2, Address);
         elsif (X1 / 8 + 1) = (X2 / 8) then
            -- X1 and X2 are adjectent bytes
            VGA_Set_Bitmask(Left_Mask);
            Draw_Strip(Y1, Y2, Address);
            VGA_Set_Bitmask(Right_Mask);
            Draw_Strip(Y1, Y2, Address+1);
         else
            -- X1 and X2 have full bytes in between
            VGA_Set_Bitmask(Left_Mask);
            Draw_Strip(Y1, Y2, Address);
            VGA_Set_Bitmask(16#FF#);
            X1 := X1 - X1 mod 8 + 8;
            for I in (X1 / 8)..(X2 / 8 - 1) loop
               Address := Address + 1;
               Draw_Strip(Y1, Y2, Address);
            end loop;
            VGA_Set_Bitmask(Right_Mask);
            Draw_Strip(Y1, Y2, Address+1);
         end if;
         VGA_Set_Write_Plane(0);
      end if;
   end Fill_Box;

   --------------------
   --  Write String  --
   --------------------
   procedure Write_String(X_Pos : in Horizontal_Location;
                          Y_Pos : in Vertical_Location;
                          Color : in Screen_Color;
                          Text  : in String) is

      procedure Plot_Line(X       : in Horizontal_Location;
                          Line    : in Integer;
                          Address : in unsigned_32;
                          Text    : in String) is

         Mask      : unsigned_8;
         Count     : Natural          := 1;
         Register  : unsigned_16      := 0;
         Offset    : constant Integer := X_Pos mod 8;
         Shift_Val : constant Integer := 8 - Offset;
         Screen    : unsigned_32      := Address +
           unsigned_32(Line) * Bytes_Per_Line;
      begin
         loop
            if X + Count * 8 + 8 <= Horizontal_Maximum then
               Register := Register and 16#FF00#;
               Register := Register or
                 unsigned_16(System_Font(Character'Pos(Text(Count)))(Line));
               Register := Rotate_Right(Register, Offset);
               Mask := unsigned_8(Register and 16#00FF#);
               VGA_Set_Bitmask(Mask);
               VGA_Put_NS_Byte(Screen, 16#FF#);
               Register := Shift_Right(Register, Shift_Val);
            end if;
            Screen   := Screen + 1;
            Count    := Count + 1;
            exit when Count > Text'Length;
         end loop;
         if X + Count * 8 + 8 <= Horizontal_Maximum then
            Register := Rotate_Right(Register, Offset);
            Mask := unsigned_8(Register and 16#00FF#);
            VGA_Set_Bitmask(Mask);
            VGA_Put_Byte(Screen, 16#FF#);
         end if;
      end Plot_Line;
      pragma Inline(Plot_Line);

      Address : unsigned_32;

   begin
      Check_Proper_Mode;
      VGA_Set_Color(Color);
      VGA_Set_Screen_Selector;
      VGA_Set_Write_Plane(VGA_All_Planes);
      Address := Calculate_Byte_Address(X_Pos, Y_Pos);
      for Y in 0..Font_Max_Height loop
         if Y_Pos + Y <= Vertical_Maximum then
            Plot_Line(X_Pos, Y, Address, Text);
         end if;
      end loop;
      VGA_Set_Write_Plane(0);
   end;

   --------------------
   --  Read_Palette  --
   --------------------
   function Read_All_Palette return Palette is
      Result : Palette;
      Regs   : go32_dpmi_registers;
      Temp   : char_array(0..Screen_Color'Pos(Screen_Color'Last));
   begin
      Check_Proper_Mode;
      if Current_Go32_Info_Block.size_of_transfer_buffer < Palette'Length then
         raise Storage_Error;
      end if;
      Regs.ax := 16#1009#;
      Regs.es := unsigned_16(
        Shift_Right(Current_Go32_Info_Block.linear_address_of_transfer_buffer,
        4) and 16#FFFF#);
      Regs.dx := unsigned_16(
        Current_Go32_Info_Block.linear_address_of_transfer_buffer and 16#0F#);
      Regs.ss := 0; Regs.sp := 0; Regs.flags := 0;
      go32_dpmi_simulate_int(BIOS_Video_Interrupt, Regs);
      dosmemget(
        Offset => Current_Go32_Info_Block.linear_address_of_transfer_buffer,
        Length => Temp'Length,
        Buffer => Temp);
      for I in Palette'Range loop
         Result(I) := Character'Pos(To_Ada(Temp(Screen_Color'Pos(I))));
      end loop;
      return Result;
   end Read_All_Palette;

   ---------------------
   --  Write Palette  --
   ---------------------
   procedure Write_All_Palette(P : in Palette) is
      Regs   : go32_dpmi_registers;
      Temp   : char_array(0..Screen_Color'Pos(Screen_Color'Last));
   begin
      Check_Proper_Mode;
      if Current_Go32_Info_Block.size_of_transfer_buffer < Palette'Length then
         raise Storage_Error;
      end if;
      for I in Screen_Color'Range loop
         Temp(Screen_Color'Pos(I)) := To_C(Character'Val(Integer(P(I))));
      end loop;
      dosmemput(
        Offset => Current_Go32_Info_Block.linear_address_of_transfer_buffer,
        Length => Temp'Length,
        Buffer => Temp);
      Regs.ax := 16#1002#;
      Regs.es := unsigned_16(
        Shift_Right(Current_Go32_Info_Block.linear_address_of_transfer_buffer,
        4) and 16#FFFF#);
      Regs.dx := unsigned_16(
        Current_Go32_Info_Block.linear_address_of_transfer_buffer and 16#0F#);
      Regs.ss := 0; Regs.sp := 0; Regs.flags := 0;
      go32_dpmi_simulate_int(BIOS_Video_Interrupt, Regs);
   end Write_All_Palette;

   ------------------------
   --  Set Palette Value --
   ------------------------
   procedure Set_Palette(P : in out Palette;
                         C : in     Screen_Color;
                         V : in     Color_Value) is
   begin
      Check_Proper_Mode;
      P(C) := V;
   end Set_Palette;

   ------------------------
   --  Get Palette Value --
   ------------------------
   function Get_Palette( P : in Palette;
                         C : in Screen_Color)
                         return Color_Value is
   begin
      Check_Proper_Mode;
      Return Color_Value(P(C));
   end Get_Palette;

   -----------------------------
   --  Copy Screen To Buffer  --
   -----------------------------
   procedure Get_Buffer(Buffer : in out Screen_Buffer;
                        Left   : in     Horizontal_Location;
                        Top    : in     Vertical_Location) is
      Bits_Left : Integer;
      Register  : unsigned_16;
      Address   : unsigned_32;
      Count     : Integer := 1;
      Offset    : constant Natural := Left mod 8;
      Rest_Bits : constant Natural := 8 - Offset;
   begin
      Check_Proper_Mode;
      if Left + Buffer.Width > Horizontal_Maximum or
        Top + Buffer.Height > Vertical_Maximum then
         raise Screen_Buffer_Error;
      else
         VGA_Set_Bitmask(16#FF#);
         VGA_Set_Screen_Selector;
         for Plane in reverse 0..3 loop
            VGA_Set_Read_Plane(unsigned_8(Plane));
            for Line in Top..Top + Buffer.Height - 1 loop
               Address := Calculate_Byte_Address(Left, Line);
               Register := Shift_Left(
                 unsigned_16(VGA_Get_NS_Byte(Address)), 8);
               Address := Address + 1;
               Register := Register or unsigned_16(VGA_Get_NS_Byte(Address));
               Address := Address + 1;
               Register := Shift_Left(Register, Offset);
               Buffer.Data(Count) := unsigned_8(Shift_Right(Register, 8));
               Count := Count + 1;
               Bits_Left := Buffer.Width - 8;
               While Bits_Left > 0 loop
                  Register := Shift_Left(Register, Rest_Bits);
                  Register := Register and 16#FF00#;
                  Register := Register or
                    unsigned_16(VGA_Get_NS_Byte(Address));
                  Address := Address + 1;
                  Register := Shift_Left(Register, Offset);
                  Buffer.Data(Count) := unsigned_8(Shift_Right(Register, 8));
                  Count := Count + 1;
                  Bits_Left := Bits_Left - 8;
               end loop;
            end loop;
         end loop;
      end if;
   end Get_Buffer;

   -----------------------------
   --  Copy Buffer To Screen  --
   -----------------------------
   procedure Put_Buffer(Buffer : in Screen_Buffer;
                        Left   : in Horizontal_Location;
                        Top    : in Vertical_Location;
                        Rule   : in Put_Rule_Type := Put_Force) is
      Bits_Left   : Integer;
      Left_Mask   : unsigned_8;
      Right_Mask  : unsigned_8;
      Register    : unsigned_16;
      Address     : unsigned_32;
      Count       : Integer := 1;
      Left_Offset : constant Integer := Left mod 8;
      Rest_Value  : constant Integer := 8 - Left_Offset;
   begin
      Check_Proper_Mode;
      if Left + Buffer.Width > Horizontal_Maximum or
        Top + Buffer.Height > Vertical_Maximum then
         raise Screen_Buffer_Error;
      else
         VGA_Set_Put_Rule(Rule);
         VGA_Set_Screen_Selector;
         Left_Mask  := Shift_Right(16#FF#, Left_Offset);
         Right_Mask := unsigned_8(Shift_Right(16#FF00#,
           ((Left + Buffer.Width - 1) mod 8) + 1) and 16#00FF#);
         if Left / 8 = (Left + Buffer.Width - 1) / 8 then
            Left_Mask := Left_Mask and Right_Mask;
         end if;
         for Plane in reverse 0..3 loop
            VGA_Set_Sequential_Plane(Shift_Left(unsigned_8(1), Plane));
            for Line in Top..Top + Buffer.Height - 1 loop
               Address := Calculate_Byte_Address(Left, Line);
               Register := unsigned_16(Buffer.Data(Count));
               Count := Count + 1;
               Register := Rotate_Right(Register, Left_Offset);
               VGA_Set_Bitmask(Left_Mask);
               VGA_Put_NS_Byte(Address, unsigned_8(Register and 16#00FF#));
               Address := Address + 1;
               Bits_Left := Buffer.Width - Rest_Value;
               VGA_Set_Bitmask(16#FF#);
               While Bits_Left >= 8 loop
                  Register := Shift_Right(Register, Rest_Value);
                  Register := Register and 16#FF00#;
                  Register := Register or unsigned_16(Buffer.Data(Count));
                  Count := Count + 1;
                  Bits_Left := Bits_Left - 8;
                  Register := Rotate_Right(Register, Left_Offset);
                  VGA_Put_NS_Byte(Address, unsigned_8(Register and 16#00FF#));
                  Address := Address + 1;
               end loop;
               VGA_Set_Bitmask(Right_Mask);
               if Bits_Left > 0 then
                  if Bits_Left > Left_Offset then
                     Register := Shift_Right(Register, Rest_Value);
                     Register := Register and 16#FF00#;
                     Register := Register or unsigned_16(Buffer.Data(Count));
                     Count := Count + 1;
                     Register := Rotate_Right(Register, Left_Offset);
                     VGA_Put_NS_Byte(Address,
                       unsigned_8(Register and 16#00FF#));
                  else
                     VGA_Put_NS_Byte(Address,
                       unsigned_8(Shift_Right(Register, 8) and 16#00FF#));
                  end if;
               end if;
            end loop;
         end loop;
         VGA_Set_Put_Rule(Put_Force);
         VGA_Set_Sequential_Plane(16#0F#);
      end if;
   end Put_Buffer;

end VGA_Graphics;