-----------------------------------------------------------------------
--
--  File:        vgagraph.ads
--  Description: Basic VGA graphics
--  Rev:         0.4
--  Date:        16/02/96
--  Author:      Jerry van Dijk
--  Mail:        jerry@jvdsys.nextjk.stuyts.nl
--
--  Copyright (c) Jerry van Dijk, 1996
--  Forelstraat 211
--  2037 KV  HAARLEM
--  THE NETHERLANDS
--  tel int + 31 23 540 1052
--
--  Permission granted to use for any purpose, provided this copyright
--  remains attached and unmodified.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
--  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-----------------------------------------------------------------------
--
--  This package provides basic VGA graphics capabilities for GNAT/DOS.
--  It is currently based on GNAT 3.01 / DJGPP 1.12m4 running under
--  both vcpi and dpmi. The package does not depend on any djgpp
--  graphics functions as it is /not/ a C library binding but a new
--  implementation written completely in Ada95 itself.
--  As it was written as a learning Ada95 program it is in its present
--  form probably not very portable to other enviroments.
--  The choice of what functionality to provide was based the expected
--  needs of a user-interface package (probably my next Ada95 program,
--  incorporating the new OO features).
--
-----------------------------------------------------------------------

with Interfaces;       use Interfaces;
with Interfaces.C;     use Interfaces.C;
with Ada.Finalization; use Ada.Finalization;

package VGA_Graphics is

-------------------
--  PUBLIC PART  --
-------------------

   --------------------------
   --  Package Exceptions  --
   --------------------------
   No_VGA_Error        : exception;       -- Couldn't switch to VGA mode
   Not_In_VGA_Error    : exception;       -- Executing VGA command in Text mode
   No_TXT_Mode_Error   : exception;       -- Couldn't return to Text mode
   Screen_Buffer_Error : exception;       -- Buffer exceeds screen size

   -----------------
   --  Font Size  --
   -----------------
   Font_Width  : constant Integer :=  8;
   Font_Height : constant Integer := 14;

   -------------------
   --  Screen Size  --
   -------------------
   Vertical_Maximum   : constant := 479;  -- Maximum line number
   Horizontal_Maximum : constant := 639;  -- Maximum column number

   subtype Vertical_Location   is Natural range 0..Vertical_Maximum;
   subtype Horizontal_Location is Natural range 0..Horizontal_Maximum;

   ---------------------
   --  Color Palette  --
   ---------------------
   type Screen_Color is (Black, Blue, Green, Cyan, Red, Magenta, Brown,
                         Light_Gray, Dark_Gray, Light_Blue, Light_Green,
                         Light_Cyan, Light_Red, Light_Magenta, Yellow,
                         White, Border);

   type Palette is private;
   subtype Color_Value is unsigned_8;

   ---------------------
   --  Screen Buffer  --
   ---------------------
   type Put_Rule_Type is (Put_Force, Put_And, Put_Or, Put_Xor);
   for Put_Rule_Type use (Put_Force =>  0,
                          Put_And   =>  8,
                          Put_Or    => 16,
                          Put_Xor   => 24);

   type Screen_Buffer(Buf_Width  : Horizontal_Location;
                      Buf_Height : Vertical_Location) is limited private;

   --------------------------
   --  Switch To VGA Mode  --
   --------------------------
   procedure VGA_Mode;

   ----------------------------
   --  Return From VGA Mode  --
   ----------------------------
   procedure TXT_Mode;

   --------------------
   --  Clear Screen  --
   --------------------
   procedure Clear_Screen(Color : in Screen_Color := Black);

   ---------------------------
   --  Put Pixel On Screen  --
   ---------------------------
   procedure Put_Pixel(X_Pos : in Horizontal_Location;
                       Y_Pos : in Vertical_Location;
                       Color : in Screen_Color);

   -----------------------
   --  Get Pixel Color  --
   -----------------------
   function Get_Pixel(X_Pos : in Horizontal_Location;
                      Y_Pos : in Vertical_Location)
                      return Screen_Color;

   ---------------------------
   --  Write Vertical Line  --
   ---------------------------
   procedure Vertical_Line(X_Pos  : in Horizontal_Location;
                           Top    : in Vertical_Location;
                           Bottom : in Vertical_Location;
                           Color  : in Screen_Color);

   -----------------------------
   --  Write Horizontal Line  --
   -----------------------------
   procedure Horizontal_Line(Left  : in Horizontal_Location;
                             Right : in Horizontal_Location;
                             Y_Pos : in Vertical_Location;
                             Color : in Screen_Color);

   -------------------
   --  Draw a line  --
   -------------------
   procedure Draw_Line(Left   : in Horizontal_Location;
                       Top    : in Vertical_Location;
                       Right  : in Horizontal_Location;
                       Bottom : in Vertical_Location;
                       Color  : in Screen_Color);

   ---------------------
   --  Draw a circle  --
   ---------------------
   procedure Draw_Circle(X_Pos  : in Horizontal_Location;
                         Y_Pos  : in Vertical_Location;
                         Radius : in Positive;
                         Color  : in Screen_Color);

   ------------------
   --  Fill a Box  --
   ------------------
   procedure Fill_Box(Left   : in Horizontal_Location;
                      Top    : in Vertical_Location;
                      Right  : in Horizontal_Location;
                      Bottom : in Vertical_Location;
                      Color  : in Screen_Color);

   --------------------
   --  Write String  --
   --------------------
   procedure Write_String(X_Pos : in Horizontal_Location;
                          Y_Pos : in Vertical_Location;
                          Color : in Screen_Color;
                          Text  : in String);

   --------------------
   --  Read_Palette  --
   --------------------
   function Read_All_Palette return Palette;

   ---------------------
   --  Write Palette  --
   ---------------------
   procedure Write_All_Palette(P : in Palette);

   ------------------------
   --  Set Palette Value --
   ------------------------
   procedure Set_Palette(P : in out Palette;
                         C : in     Screen_Color;
                         V : in     Color_Value);

   ------------------------
   --  Get Palette Value --
   ------------------------
   function Get_Palette(P : in Palette;
                        C : in Screen_Color)
                        return Color_Value;

   ----------------------------------
   --  Copy Screen Area To Buffer  --
   ----------------------------------
   procedure Get_Buffer(Buffer : in out Screen_Buffer;
                        Left   : in     Horizontal_Location;
                        Top    : in     Vertical_Location);

   ----------------------------
   --  Copy Buffer To Screen --
   ----------------------------
   procedure Put_Buffer(Buffer : in Screen_Buffer;
                        Left   : in Horizontal_Location;
                        Top    : in Vertical_Location;
                        Rule   : in Put_Rule_Type := Put_Force);

private

--------------------
--  PRIVATE PART  --
--------------------

   ---------------------
   --  Screen Buffer  --
   ---------------------
   type Planes is array(Natural range <>) of unsigned_8;
   type Planes_Access is access Planes;

   type Screen_Buffer(Buf_Width  : Horizontal_Location;
                      Buf_Height : Vertical_Location)
                      is new Limited_Controlled with
      record
         Width  : Horizontal_Location := Buf_Width;
         Height : Vertical_Location   := Buf_Height;
         Data   : Planes_Access;
      end record;

   procedure Initialize(B : in out Screen_Buffer);
   procedure Finalize(B : in out Screen_Buffer);

   ---------------------
   --  Color Palette  --
   ---------------------
   type Palette is array (Screen_Color range Black..Border) of Color_Value;

end VGA_Graphics;