--  The Ada 95 Booch Components (Version 0.4q)
--  (C) Copyright 1987-1994 Grady Booch.
--  (C) Copyright 1994-1996 Grady Booch and David Weller.
--
-- The Ada 95 Booch Components is a collection of free software components.
-- You can redistribute it and/or modify it under terms of the Ada
-- Community License.
--

with Ada.Finalization; use Ada.Finalization;
generic
    type Item is private;
package Bc.Queues is

    type Queue is abstract new Controlled with private;

	 -- A sequence in which items may be added from one end and removed from
	 -- the opposite end.  This class is abstract and serves only to enforce
	 -- the interfaces among classes.

	 -- Operations of equality, inequality, and assignment are "deep" for
	 -- all Queue forms

    type Item_Ptr is access all Item;

    procedure Clear   (Obj : in out Queue) is abstract;
	 -- Empty the Queue of all items

    procedure Append  (Obj : in out Queue; Elem : Item) is abstract;
	 -- Add the item to the back of the Queue; the Item itself is copied

    procedure Pop     (Obj : in out Queue) is abstract;
	 -- Remove the Item from the front of the Queue

    procedure Remove  (Obj : in out Queue; From : Natural) is abstract;
	 -- Remove the Item at the given Index (balking operation)

    function Length   (Obj : in Queue) return Natural is abstract;
	 -- Returns total items in the Queue

    function Is_Empty (Obj : in Queue) return Boolean is abstract;
	 -- Returns true iff no items are in the queue

    function Front    (Obj : in Queue) return Item is abstract;
	 -- Return the item at the front of the Queue; the Item is _not_ removed

    function Location (Obj : in Queue; Elem : in Item) return Natural 
		 is abstract;
	 -- Returns the Index number where the Item is found.  A zero is
	 -- returned on failure

    function "=" (Left, Right : access Queue'Class) return Boolean;

    procedure Copy (From : access Queue'Class; To : access Queue'Class);
	 -- This operation MUST be called for dissimilar Queues in place of
	 -- assignment.

    type Iterator (Q : access Queue'Class) is limited private;
	 -- An active iterator.  User must explicitly advance and check the 
	 -- iterator.

    procedure Reset (Obj : in out Iterator);
	 -- Resets iterator to point to first Item in Queue

    procedure Next (Obj : in out Iterator);
	 -- Advances to next Item in Queue

    function Is_Done (Obj : Iterator) return Boolean;
	 -- Returns true iff the Iterator cannot advance further

    function Current_Item (Obj : Iterator) return Item_Ptr;
	 -- Returns a reference to the Item's location in memory.  Use with
	 -- caution.

    type Passive_Iterator (Q : access Queue'Class) is limited private;
	 -- A passive iterator.  Automatically advances over Queue when the
	 -- 'Apply' procedure is called

    generic
       with procedure Apply_Method(Elem : in out Item; OK : out Boolean);
    procedure Apply (Obj : in out Passive_Iterator; Success : out Boolean);
	 -- This generic must be instantiated for proper operation of a Passive
	 -- Iterator.

private

    type Queue is abstract new Controlled with null record;

    type Iterator (Q : access Queue'Class) is limited record
       Index : Natural := 1;
    end record;

    type Passive_Iterator (Q : access Queue'Class) is limited record
       Success : Boolean;
    end record;

    function Item_At (Obj : Queue; Index : Natural) return Item;
    function Item_At (Obj : Queue; Index : Natural) return Item_Ptr;
    procedure Purge (Obj : in out Queue);
    procedure Add (Obj : in out Queue; Elem : in out Item);
    function Cardinality (Obj : Queue) return Integer;

end Bc.Queues;
