WITH Ada.Text_IO;
WITH Screen; 
PACKAGE BODY Spiders IS
------------------------------------------------------------------
--|
--| Body of package providing procedures to emulate multiple
--| spiders. The spiders can move around
--| the screen drawing simple patterns. Monochrome version.
--|
--| Author: John Dalbey, Cal Poly San Luis Obispo, 1992
--| Adapted by: Michael B. Feldman, The George Washington University
--| Last Modified: July 1995
--|
------------------------------------------------------------------

  TYPE DirectionSymbols IS ARRAY (Directions) OF character;
  TYPE Palette          IS ARRAY (ScreenColors) OF character;

  DebugFlag    : Boolean := False;             -- Is single stepping on?
  RowHi        : Screen.Depth := 22;          -- room upper boundary for row
  ColHi        : Screen.Width := 40;          -- room upper boundary for column

  Spidersym    : CONSTANT character := '*';   -- asterisk
  LoBound      : CONSTANT integer := 1;       -- room lower boundary
  WindowOffset : CONSTANT integer := 20;
  ColorSymbols : CONSTANT Palette := ('+','X','O','.'); 
                                              -- ASCII symbols for color

  PROCEDURE DrawRoom  IS
  BEGIN
    Screen.ClearScreen;
    Screen.MoveCursor (Row => 1, Column => 1);
    -- Top Bar
    Ada.Text_IO.Put ("                    ");
    Ada.Text_IO.Put ("----------------------------------------- ");
    Ada.Text_IO.New_Line;
    FOR I in 1..21 LOOP
    Ada.Text_IO.Put ("                   |");
    Ada.Text_IO.Put (". . . . . . . . . . . . . . . . . . . . .|");
    Ada.Text_IO.New_Line;
    END LOOP;
    Ada.Text_IO.Put ("                    ");
    Ada.Text_IO.Put ("----------------------------------------- ");
  END DrawRoom;

  PROCEDURE ChgColor (Which : IN OUT Spider; NewColor : ScreenColors) IS
  -- Change the color the spider is using.
  BEGIN
    Which.Ink := NewColor;
  END ChgColor;

  PROCEDURE ShowSpider (Which: Spider) IS
  -- Show the spider symbol
  BEGIN
    Screen.MoveCursor (Row => Which.CurrentRow+1, 
                       Column => Which.CurrentColumn+WindowOffset);
    Ada.Text_IO.Put (SpiderSym);
    Screen.MoveCursor ( Row => 3,Column => 3);  -- HIdecursor
  END ShowSpider;

  PROCEDURE Start (Which: IN OUT Spider) IS
  -- Create a fixed size room and reset the spider.
  BEGIN
    Which.CurrentColumn := 21;
    Which.CurrentRow := 11;
    Which.Heading := North;
    Which.Ink := Green;
    ShowSpider (Which);
  END Start;

  -- Color commands
  PROCEDURE Blue (Which: IN OUT Spider) IS
  BEGIN
    ChgColor (Which, Blue);
  END Blue;

  PROCEDURE Green (Which: IN OUT Spider) IS
  BEGIN
    ChgColor (Which, Green);
  END Green;

  PROCEDURE Red (Which: IN OUT Spider) IS
  BEGIN
    ChgColor (Which, Red);
  END Red;

  PROCEDURE Black (Which: IN OUT Spider) IS
  BEGIN
    ChgColor (Which, Black);
  END Black;

  PROCEDURE Face (Which: IN OUT Spider; WhichWay: IN Directions) IS
  -- Turn to the desired direction
  BEGIN
    Which.Heading := WhichWay;
  END Face;

  FUNCTION IsFacing (Which: Spider) RETURN Directions IS
  -- Return current direction
  BEGIN
    RETURN Which.Heading;
  END IsFacing;

  PROCEDURE Jump (Which: IN OUT Spider; HowMany: IN Positive) IS
    AnyThing: character;
  BEGIN
    -- put a block down where spider is standing
    Screen.MoveCursor(Row => Which.CurrentRow+1,
                      Column => Which.CurrentColumn+WindowOffset);
    Ada.Text_IO.Put (ColorSymbols (Which.Ink) );
    
    -- If out of bounds raise exception.
    IF NearWall (Which, HowMany) THEN
      Quit;
      RAISE Hit_the_Wall;
    END IF;

    -- change the location coordinates
    CASE Which.Heading IS
      WHEN North => 
        Which.CurrentRow := Which.CurrentRow - HowMany;
      WHEN East  => 
        Which.CurrentColumn := Which.CurrentColumn + 2*HowMany;
      WHEN South => 
        Which.CurrentRow := Which.CurrentRow + HowMany;
      WHEN West  => 
        Which.CurrentColumn := Which.CurrentColumn - 2*HowMany;
    END CASE;

    -- draw the spider in her new location
    ShowSpider (Which);
    IF Debugging = On THEN   -- if debug mode, wait for user to press RETURN
      Ada.Text_IO.Skip_Line;
    ELSE
      DELAY 0.5;
    END IF;
  END Jump;

  PROCEDURE Step (Which : IN OUT Spider) IS
  BEGIN
    Jump (Which, 1);
  END Step;

  PROCEDURE Right (Which: IN OUT Spider) IS
  -- Turn to the right command.
  BEGIN
    IF Which.Heading = Directions'Last THEN
      Which.Heading := Directions'First;
    ELSE
      Which.Heading := Directions'Succ (Which.Heading);
    END IF;
  END Right; 

  FUNCTION  NearWall (Which: Spider; HowMany: Positive) RETURN Boolean IS
  -- RETURN True if spider is near a wall
  BEGIN
    -- Check for out of bounds
    CASE Which.Heading IS
      WHEN North => 
        RETURN Which.CurrentRow - (HowMany - 1) <= LoBound;
      WHEN East  => 
        RETURN Which.CurrentColumn + 2*(HowMany - 1) >= ColHi;
      WHEN South => 
        RETURN  Which.CurrentRow + (HowMany - 1) >= RowHi;
      WHEN West  => 
        RETURN Which.CurrentColumn - 2*(HowMany - 1) <= LoBound;
    END CASE;
  END NearWall;

  FUNCTION AtWall (Which : Spider) RETURN Boolean IS
  BEGIN
    RETURN NearWall (Which, 1);
  END AtWall;

  PROCEDURE Quit IS
  -- Quit command.
  BEGIN
    Screen.MoveCursor(Row => 23,Column => 1);
  END Quit;

  PROCEDURE Debug (Setting: Switch) is
  -- Toggle debugging mode
  BEGIN
    IF Setting = ON THEN
      DebugFlag := true;
      Screen.MoveCursor (Row => 10,Column => 1);
      Ada.Text_IO.Put ("-- DEBUG ON -- ");
      Ada.Text_IO.New_Line;
      Ada.Text_IO.Put ("  Press Enter");
    ELSE
      DebugFlag := false;
      Screen.MoveCursor (Row => 10,Column => 1);
      Ada.Text_IO.Put ("               ");
      Ada.Text_IO.New_Line;
      Ada.Text_IO.Put ("             ");
    END IF;
  END Debug;

  FUNCTION Debugging RETURN Switch IS 
  BEGIN
    IF DebugFlag THEN
      RETURN On;
    ELSE
      RETURN Off;
    END IF;
  END Debugging;

END Spiders;
